
#include "mexutil.h"
#include "nn.h"
#include <string.h>

BITMAP *convert_bitmap(const mxArray *A) {
  if (mxGetNumberOfDimensions(A) != 3) { mexErrMsgTxt("dims != 3"); }
  if (mxGetDimensions(A)[2] != 3) { mexErrMsgTxt("color channels != 3"); }

  int h = mxGetDimensions(A)[0];
  int w = mxGetDimensions(A)[1];

  if (mxIsUint8(A)) {
    unsigned char *data = (unsigned char *) mxGetData(A);
    BITMAP *ans = create_bitmap(w, h);
    for (int y = 0; y < h; y++) {
      int *row = (int *) ans->line[y];
      for (int x = 0; x < w; x++) {
        int r = data[y+x*h];
        int g = data[(y+x*h)+w*h];
        int b = data[(y+x*h)+2*w*h];
        if (r < 0) { r = 0; }
        if (g < 0) { g = 0; }
        if (b < 0) { b = 0; }
        if (r > 255) { r = 255; }
        if (g > 255) { g = 255; }
        if (b > 255) { b = 255; }
        row[x] = r|(g<<8)|(b<<16);
      }
    }
    return ans;
  } else if (mxIsSingle(A)) {
    float *data = (float *) mxGetData(A);
    BITMAP *ans = create_bitmap(w, h);
    for (int y = 0; y < h; y++) {
      int *row = (int *) ans->line[y];
      for (int x = 0; x < w; x++) {
        float r = data[y+x*h];
        float g = data[(y+x*h)+w*h];
        float b = data[(y+x*h)+2*w*h];
        if (r < 0) { r = 0; }
        if (g < 0) { g = 0; }
        if (b < 0) { b = 0; }
        if (r > 1) { r = 1; }
        if (g > 1) { g = 1; }
        if (b > 1) { b = 1; }
        row[x] = int(r*255)|(int(g*255)<<8)|(int(b*255)<<16);
      }
    }
    return ans;
  } else if (mxIsDouble(A)) {
    double *data = (double *) mxGetData(A);
    BITMAP *ans = create_bitmap(w, h);
    for (int y = 0; y < h; y++) {
      int *row = (int *) ans->line[y];
      for (int x = 0; x < w; x++) {
        double r = data[y+x*h];
        double g = data[(y+x*h)+w*h];
        double b = data[(y+x*h)+2*w*h];
        if (r < 0) { r = 0; }
        if (g < 0) { g = 0; }
        if (b < 0) { b = 0; }
        if (r > 1) { r = 1; }
        if (g > 1) { g = 1; }
        if (b > 1) { b = 1; }
        row[x] = int(r*255)|(int(g*255)<<8)|(int(b*255)<<16);
      }
    }
    return ans;
  } else {
    mexErrMsgTxt("bitmap not uint8, single, or double");
  }
}

BITMAP *convert_bitmapf(const mxArray *A) {
  if (!(mxGetNumberOfDimensions(A) == 2 ||
       (mxGetNumberOfDimensions(A) == 3 && mxGetDimensions(A)[2] == 3))) { mexErrMsgTxt("float bitmap doesn't have 2 dims or 3 dims with 3 channels"); }

  int h = mxGetDimensions(A)[0];
  int w = mxGetDimensions(A)[1];

  if (mxIsUint8(A)) {
    unsigned char *data = (unsigned char *) mxGetData(A);
    BITMAP *ans = create_bitmap(w, h);
    for (int y = 0; y < h; y++) {
      float *row = (float *) ans->line[y];
      for (int x = 0; x < w; x++) {
        row[x] = data[y+x*h];
      }
    }
    return ans;
  } else if (mxIsDouble(A)) {
    double *data = (double *) mxGetData(A);
    BITMAP *ans = create_bitmap(w, h);
    for (int y = 0; y < h; y++) {
      float *row = (float *) ans->line[y];
      for (int x = 0; x < w; x++) {
        row[x] = data[y+x*h];
      }
    }
    return ans;
  } else {
    mexErrMsgTxt("float bitmap not uint8 or double");
  }
}

BITMAP *convert_field(Params *p, const mxArray *A, int bw, int bh, int &nclip, int trim_patch) {
  nclip = 0;
  int ndims = mxGetNumberOfDimensions(A);
  if (ndims != 3) {
    char buf[256];
    if (ndims == 1) {
      sprintf(buf, "field dims != 3 (%d 1d array where nn field expected)", mxGetDimensions(A)[0]);
    } else if (ndims == 2) {
      sprintf(buf, "field dims != 3 (%dx%d 2d array where nn field expected)", mxGetDimensions(A)[0], mxGetDimensions(A)[1]);
    } else {
      sprintf(buf, "field dims != 3 (%d dimension array where nn field expected)", ndims);
    }
    mexErrMsgTxt(buf);
  }
  if (mxGetDimensions(A)[2] != 3) { mexErrMsgTxt("field channels != 3"); }
  if (!mxIsInt32(A)) { mexErrMsgTxt("field is not int32"); }

  int h = mxGetDimensions(A)[0];
  int w = mxGetDimensions(A)[1];
  int bew = trim_patch ? (bw-p->patch_w+1): bw;
  int beh = trim_patch ? (bh-p->patch_w+1): bh;

  unsigned int *data = (unsigned int *) mxGetData(A);
  BITMAP *ann = create_bitmap(w, h);
  //annd = create_bitmap(w, h);
  for (int y = 0; y < h; y++) {
    int *ann_row = (int *) ann->line[y];
    //int *annd_row = (int *) annd->line[y];
    for (int x = 0; x < w; x++) {
      int xp = data[y+x*h];
      int yp = data[(y+x*h)+w*h];
      if ((unsigned) xp >= (unsigned) bew ||
          (unsigned) yp >= (unsigned) beh) {
        nclip++;
        if (xp < 0) { xp = 0; }
        if (yp < 0) { yp = 0; }
        if (xp >= bew) { xp = bew-1; }
        if (yp >= beh) { yp = beh-1; }
      }

      //int dp = data[(y+x*h)+2*w*h];
      ann_row[x] = XY_TO_INT(xp, yp);
      //annd_row[x] = dp;
    }
  }
  return ann;
}

BITMAP *convert_winsize_field(Params *p, const mxArray *A, int w, int h) {
  int ndims = mxGetNumberOfDimensions(A);
  if (!( (ndims == 2)||((ndims == 3)&&(mxGetDimensions(A)[2] == 2))&&(w==mxGetDimensions(A)[1])&&(h==mxGetDimensions(A)[0]) )) {
	char buf[256];
	//sprintf(buf, "ndims=%d, mxGetDim(A)=[%d %d %d], [2 w h]=[2 %d %d] \n", ndims,mxGetDimensions(A)[2],mxGetDimensions(A)[1],mxGetDimensions(A)[0],h,w);
    sprintf(buf, "winsize field dims=%d (a [%dx%d] or [%dx%dx2] field of window size is expected)", ndims, h, w, h, w);
    mexErrMsgTxt(buf);
  }
  if (!mxIsInt32(A)) { mexErrMsgTxt("winsize field is not int32"); }

  unsigned int *data = (unsigned int *) mxGetData(A);
  BITMAP *awsz = create_bitmap(w, h);
  char buf[256];
  if (ndims == 2) {
	for (int y = 0; y < h; y++) {
	  int *awsz_row = (int *) awsz->line[y];
	  for (int x = 0; x < w; x++) {
		int wp = data[y+x*h];
		awsz_row[x] = XY_TO_INT(wp, wp);
	  }
	}
  } else if ((ndims == 3)&&(mxGetDimensions(A)[2] == 2)) {
	for (int y = 0; y < h; y++) {
	  int *awsz_row = (int *) awsz->line[y];
	  for (int x = 0; x < w; x++) {
		int wp = data[y+x*h];
		int hp = data[(y+x*h)+w*h];
		awsz_row[x] = XY_TO_INT(wp, hp);
	  }
    }
  }
  return awsz;
}

mxArray *bitmap_to_array(BITMAP *a) {
  mwSize dims[3] = { a->h, a->w, 3 };
  mxArray *ans = mxCreateNumericArray(3, dims, mxUINT8_CLASS, mxREAL);
  unsigned char *data = (unsigned char *) mxGetData(ans);
  unsigned char *rchan = &data[0];
  unsigned char *gchan = &data[a->w*a->h];
  unsigned char *bchan = &data[2*a->w*a->h];
  for (int y = 0; y < a->h; y++) {
    int *row = (int *) a->line[y];
    for (int x = 0; x < a->w; x++) {
      int c = row[x];
      rchan[y+x*a->h] = c&255;
      gchan[y+x*a->h] = (c>>8)&255;
      bchan[y+x*a->h] = (c>>16);
    }
  }
  return ans;
}

int mxStringEquals(const mxArray *A, const char *s) {
  char buf[256];
  if (!mxIsChar(A)) { return 0; }
  if (mxGetString(A, buf, 255)) { return 0; }
  return strcmp(s, buf) == 0;
}
