
#ifndef _nnvec_h
#define _nnvec_h

#include "allegro_emu.h"
#include <omp.h>
#include <float.h>
#include "nn.h"

/* -------------------------------------------------------------------------
   Randomized NN algorithm, on vector input (patches, SIFT descriptors, etc)
   -------------------------------------------------------------------------
   
   Descriptors are vectors of type T and dimension n.
   Note that ACCUM must be a signed type, and must be large enough to hold the distance between any two descriptors. */

template<class T>
class VECBITMAP { public:
  T *data;
  int w, h, n;
  T *get(int x, int y) { return &data[(y*w+x)*n]; }   /* Get patch (x, y). */
  T *line_n1(int y) { return &data[y*w]; }            /* Get line y assuming n=1. */
  VECBITMAP() { }
  VECBITMAP(int w_, int h_, int n_) { w = w_; h = h_; n = n_; data = new T[w*h*n]; }
  ~VECBITMAP() { delete[] data; }
};

template<class T>
T get_maxval() { fprintf(stderr, "get_maxval of unsupported template type\n"); exit(1); }

template<> int get_maxval();
template<> float get_maxval();
template<> double get_maxval();
template<> long long get_maxval();

#include "vecpatch.h"

template<class T>
BITMAP *wrap_vecbitmap(VECBITMAP<T> *a) {
  BITMAP *ans = new BITMAP();
  ans->w = a->w;
  ans->h = a->h;
  return ans;
}

template<class T>
BITMAP *vec_init_nn(Params *p, VECBITMAP<T> *a, VECBITMAP<T> *b, BITMAP *bmask=NULL, RegionMasks *region_masks=NULL, RegionMasks *amask=NULL) {
  BITMAP *awrap = wrap_vecbitmap(a);
  BITMAP *bwrap = wrap_vecbitmap(b);

  BITMAP *ans = init_nn(p, awrap, bwrap, bmask, region_masks, amask, 0);

  delete awrap;
  delete bwrap;
  return ans;
}

template<class T>
Box get_abox_vec(Params *p, VECBITMAP<T> *a, RegionMasks *amask, int trim_patch=1) {
  BITMAP *aw = wrap_vecbitmap(a);
  Box ans = get_abox(p, aw, amask, trim_patch);
  delete aw;
  return ans;
}

template<class T, class ACCUM, int IS_MASK, int IS_WINDOW>
VECBITMAP<ACCUM> *vec_init_dist_n(Params *p, VECBITMAP<T> *a, VECBITMAP<T> *b, BITMAP *ann, BITMAP *bmask=NULL, RegionMasks *region_masks=NULL, RegionMasks *amask=NULL) {
  VECBITMAP<ACCUM> *ans = new VECBITMAP<ACCUM>(a->w, a->h, 1);
  ACCUM maxval = get_maxval<ACCUM>();
  for (int y = 0; y < ans->h; y++) {
    ACCUM *row = ans->line_n1(y);
    for (int x = 0; x < ans->w; x++) {
      row[x] = maxval;
    }
  }
  if (region_masks) {
    if (region_masks->bmp->w != a->w || region_masks->bmp->h != a->h) { fprintf(stderr, "region_masks (%dx%d) size != a (%dx%d) size\n", region_masks->bmp->w, region_masks->bmp->h, a->w, a->h); exit(1); }
    if (region_masks->bmp->w != b->w || region_masks->bmp->h != b->h) { fprintf(stderr, "region_masks (%dx%d) size != b (%dx%d) size\n", region_masks->bmp->w, region_masks->bmp->h, b->w, b->h); exit(1); }
  }

  Box box = get_abox_vec(p, a, amask, 0);
#pragma omp parallel for schedule(static, 4)
  for (int y = box.ymin; y < box.ymax; y++) {
    ACCUM *row = (ACCUM *) ans->line_n1(y);
    int *arow = amask ? (int *) amask->bmp->line[y]: NULL;
    for (int x = box.xmin; x < box.xmax; x++) {
      if (IS_MASK && amask && arow[x]) { continue; }
      int xp, yp;
      getnn(ann, x, y, xp, yp);

      if (IS_MASK && region_masks && ((int *) region_masks->bmp->line[y])[x] != ((int *) region_masks->bmp->line[yp])[xp]) {
        row[x] = maxval; continue;
      }

      T *apatch = a->get(x, y);

      if (IS_MASK && bmask && ((int *) bmask->line[yp])[xp]) { row[x] = maxval; continue; }
      T *bpatch = b->get(xp, yp);

      row[x] = vec_fast_patch_nobranch<T, ACCUM, IS_WINDOW>(apatch, bpatch, p);
      //if (x == 1 && y == 1) { printf("1, 1 => %d, %d (%d)\n", xp, yp, row[x]); }
    }
  }
  return ans;
}

template<class T, class ACCUM>
VECBITMAP<ACCUM> *vec_init_dist(Params *p, VECBITMAP<T> *a, VECBITMAP<T> *b, BITMAP *ann, BITMAP *bmask=NULL, RegionMasks *region_masks=NULL, RegionMasks *amask=NULL) {
  if (is_window(p)) {
    return vec_init_dist_n<T, ACCUM, 1, 1>(p, a, b, ann, bmask, region_masks, amask);
  } else if (amask || bmask || region_masks) {
    return vec_init_dist_n<T, ACCUM, 1, 0>(p, a, b, ann, bmask, region_masks, amask);
  } else {
    return vec_init_dist_n<T, ACCUM, 0, 0>(p, a, b, ann, bmask, region_masks, amask);
  }
}

template<class T>
int window_constraint_wrap(Params *p, VECBITMAP<T> *a, VECBITMAP<T> *b, int ax, int ay, int bx, int by) {
  BITMAP aw, bw;
  aw.w = a->w;
  aw.h = a->h;
  bw.w = b->w;
  bw.h = b->h;
  return window_constraint(p, &aw, &bw, ax, ay, bx, by);
}

template<class T, class ACCUM, int IS_MASK, int IS_WINDOW>
void vec_nn_n(Params *p, VECBITMAP<T> *a, VECBITMAP<T> *b,
            BITMAP *ann, VECBITMAP<ACCUM> *annd,
            RegionMasks *amask=NULL, BITMAP *bmask=NULL,
            int level=0, int em_iter=0, RecomposeParams *rp=NULL, int offset_iter=0, int update_type=0, int cache_b=0,
            RegionMasks *region_masks=NULL, int tiles=-1) {
  if (tiles < 0) { tiles = p->cores; }
  printf("in vec_nn_n, masks are: %p %p %p, tiles=%d, rs_max=%d\n", amask, bmask, region_masks, tiles, p->rs_max);
  Box box = get_abox_vec(p, a, amask, 0);
  int nn_iter = 0;
  for (; nn_iter < p->nn_iters; nn_iter++) {
    unsigned int iter_seed = rand();

    #pragma omp parallel num_threads(tiles)
    {
      int ithread = omp_get_thread_num();
      int xmin = box.xmin, xmax = box.xmax;
      int ymin = box.ymin + (box.ymax-box.ymin)*ithread/tiles;
      int ymax = box.ymin + (box.ymax-box.ymin)*(ithread+1)/tiles;

      int ystart = ymin, yfinal = ymax, ychange=1;
      int xstart = xmin, xfinal = xmax, xchange=1;
      if ((nn_iter + offset_iter) % 2 == 1) {
        xstart = xmax-1; xfinal = xmin-1; xchange=-1;
        ystart = ymax-1; yfinal = ymin-1; ychange=-1;
      }
      int dx = -xchange, dy = -ychange;

      int bew = b->w, beh = b->h;
      int max_mag = MAX(b->w, b->h);
      int rs_ipart = int(p->rs_iters);
      double rs_fpart = p->rs_iters - rs_ipart;
      int rs_max = p->rs_max;
      if (rs_max > max_mag) { rs_max = max_mag; }

      for (int y = ystart; y != yfinal; y += ychange) {
        ACCUM *annd_row = annd->line_n1(y);
        int *amask_row = IS_MASK ? (amask ? (int *) amask->bmp->line[y]: NULL): NULL;
        for (int x = xstart; x != xfinal; x += xchange) {
          if (IS_MASK && amask && amask_row[x]) { continue; }

          T *apatch = a->get(x, y);
          
          int src_mask = IS_MASK ? (region_masks ? ((int *) region_masks->bmp->line[y])[x]: 0): 0;

          int xbest, ybest;
          getnn(ann, x, y, xbest, ybest);
          ACCUM err = annd_row[x];
          if (!err) { continue; }

          /* Propagate */
          if (p->do_propagate) {
            /* Propagate x */
            if ((unsigned) (x+dx) < (unsigned) (ann->w)) {
              int xpp, ypp;
              getnn(ann, x+dx, y, xpp, ypp);
              xpp -= dx;

              if (!IS_WINDOW || window_constraint_wrap(p, a, b, x, y, xpp, ypp)) {
                vec_attempt_n<T, ACCUM, IS_MASK, IS_WINDOW>(err, xbest, ybest, apatch, b, xpp, ypp, bmask, region_masks, src_mask, p);
              }
            }

            /* Propagate y */
            if ((unsigned) (y+dy) < (unsigned) (ann->h)) {
              int xpp, ypp;
              getnn(ann, x, y+dy, xpp, ypp);
              ypp -= dy;

              if (!IS_WINDOW || window_constraint_wrap(p, a, b, x, y, xpp, ypp)) {
                vec_attempt_n<T, ACCUM, IS_MASK, IS_WINDOW>(err, xbest, ybest, apatch, b, xpp, ypp, bmask, region_masks, src_mask, p);
              }
            }
          }

          /* Random search */
          unsigned int seed = (x | (y<<11)) ^ iter_seed;
          seed = RANDI(seed);
          int rs_iters = 1-(seed*(1.0/(RAND_MAX-1))) < rs_fpart ? rs_ipart + 1: rs_ipart;
  //        int rs_iters = 1-random() < rs_fpart ? rs_ipart + 1: rs_ipart;

          int rs_max_curr = rs_max;
          for (int mag = rs_max_curr; mag >= p->rs_min; mag = int(mag*p->rs_ratio)) {
            for (int rs_iter = 0; rs_iter < rs_iters; rs_iter++) {
              int xmin = xbest-mag, xmax = xbest+mag;
              int ymin = ybest-mag, ymax = ybest+mag;
              xmax++;
              ymax++;
              if (xmin < 0) { xmin = 0; }
              if (ymin < 0) { ymin = 0; }
              if (xmax > bew) { xmax = bew; }
              if (ymax > beh) { ymax = beh; }

              seed = RANDI(seed);
              int xpp = xmin+seed%(xmax-xmin);
              seed = RANDI(seed);
              int ypp = ymin+seed%(ymax-ymin);
              //int xpp = xmin+rand()%(xmax-xmin);
              //int ypp = ymin+rand()%(ymax-ymin);
              if (!IS_WINDOW || window_constraint_wrap(p, a, b, x, y, xpp, ypp)) {
                vec_attempt_n<T, ACCUM, IS_MASK, IS_WINDOW>(err, xbest, ybest, apatch, b, xpp, ypp, bmask, region_masks, src_mask, p);
              }
            }
          }
          
          ((int *) ann->line[y])[x] = XY_TO_INT(xbest, ybest);
          annd_row[x] = err;
        } // x
      } // y
    } // parallel
  } // nn_iter
  printf("done vec_nn_n, %d iters, rs_max=%d\n", nn_iter, p->rs_max);
}

template<class T, class ACCUM>
void vec_nn(Params *p, VECBITMAP<T> *a, VECBITMAP<T> *b,
            BITMAP *ann, VECBITMAP<ACCUM> *annd,
            RegionMasks *amask=NULL, BITMAP *bmask=NULL,
            int level=0, int em_iter=0, RecomposeParams *rp=NULL, int offset_iter=0, int update_type=0, int cache_b=0,
            RegionMasks *region_masks=NULL, int tiles=-1) {
  if (p->algo == ALGO_CPU || p->algo == ALGO_CPUTILED) {
    if (is_window(p)) {
      printf("Running vec_nn (cputiled), using windowed and masked\n");
      vec_nn_n<T, ACCUM, 1, 1>(p, a, b, ann, annd, amask, bmask, level, em_iter, rp, offset_iter, update_type, cache_b, region_masks, tiles);
    } else if (bmask == NULL && amask == NULL && region_masks == NULL) {
      printf("Running vec_nn (cputiled), using unmasked\n");
      vec_nn_n<T, ACCUM, 0, 0>(p, a, b, ann, annd, amask, bmask, level, em_iter, rp, offset_iter, update_type, cache_b, region_masks, tiles);
    } else {
      printf("Running vec_nn (cputiled), using masked\n");
      vec_nn_n<T, ACCUM, 1, 0>(p, a, b, ann, annd, amask, bmask, level, em_iter, rp, offset_iter, update_type, cache_b, region_masks, tiles);
    }
  } else {
    fprintf(stderr, "vec_nn: algorithm %d unsupported\n", p->algo); exit(1);
  }
}

template<class T, class ACCUM, int IS_WINDOW, int HAS_MASKS>
void vec_minnn_n(Params *p, VECBITMAP<T> *a, VECBITMAP<T> *b, BITMAP *ann, VECBITMAP<ACCUM> *annd, BITMAP *ann_prev, BITMAP *bmask, int level, int em_iter, RecomposeParams *rp, RegionMasks *region_masks, RegionMasks *amask, int ntiles) {
  if (ntiles < 0) { ntiles = p->cores; }
  printf("vec_minnn: %d %d %d %d, tiles=%d\n", ann->w, ann->h, ann_prev->w, ann_prev->h, ntiles);
  if (!rp) { fprintf(stderr, "vec_minnn_n: rp is NULL\n"); exit(1); }
//  double start_t = accurate_timer();
  Box box = get_abox_vec(p, a, amask, 0);

  #pragma omp parallel for schedule(static,4) num_threads(ntiles)
  for (int y = box.ymin; y < box.ymax; y++) {
    int *amask_row = amask ? (int *) amask->bmp->line[y]: NULL;
    ACCUM *annd_row = (ACCUM *) annd->line_n1(y);
    for (int x = box.xmin; x < box.xmax; x++) {
      if (HAS_MASKS && amask && amask_row[x]) { continue; }

      ACCUM dcurrent = annd_row[x];
      int xp, yp;
      getnn(ann_prev, x, y, xp, yp);
      if ((unsigned) xp >= (unsigned) (b->w) ||
          (unsigned) yp >= (unsigned) (b->h)) { continue; }
      if (HAS_MASKS && bmask && ((int *) bmask->line[yp])[xp]) { continue; }

      //int dprev = patch_dist(p, a, x, y, b, xp, yp, dcurrent, region_masks);
      ACCUM dprev = vec_patch_dist_ab<T, ACCUM, IS_WINDOW, HAS_MASKS>(p, a, x, y, b, xp, yp, dcurrent, region_masks);
      
      if (dprev < dcurrent) {
        _putpixel32(ann, x, y, XY_TO_INT(xp, yp));
        annd_row[x] = dprev;
      }
    }
  }
//  nn_time += accurate_timer() - start_t;
  Params pcopy(*p);
  pcopy.nn_iters = rp->minnn_optp_nn_iters;
  pcopy.rs_max = rp->minnn_optp_rs_max;
  
  vec_nn<T, ACCUM>(&pcopy, a, b, ann, annd, amask, bmask, level, em_iter, rp, 0, 0, 1, region_masks, ntiles);
}

template<class T, class ACCUM>
void vec_minnn(Params *p, VECBITMAP<T> *a, VECBITMAP<T> *b, BITMAP *ann, VECBITMAP<ACCUM> *annd, BITMAP *ann_prev, BITMAP *bmask, int level, int em_iter, RecomposeParams *rp, RegionMasks *region_masks, RegionMasks *amask, int ntiles) {
  if (is_window(p)) {
    return vec_minnn_n<T, ACCUM, 1, 1>(p, a, b, ann, annd, ann_prev, bmask, level, em_iter, rp, region_masks, amask, ntiles);
  } else if (bmask || region_masks || amask) {
    return vec_minnn_n<T, ACCUM, 0, 1>(p, a, b, ann, annd, ann_prev, bmask, level, em_iter, rp, region_masks, amask, ntiles);
  } else {
    return vec_minnn_n<T, ACCUM, 0, 0>(p, a, b, ann, annd, ann_prev, bmask, level, em_iter, rp, region_masks, amask, ntiles);
  }
}

BITMAP *vecbitmap_to_bitmap(VECBITMAP<int> *a);

#define VEC_MODE_PATCH 0
#define VEC_MODE_DESC  1
#define VEC_MODE_SIM   2

BITMAP *vecwrap_init_nn(int vec_mode, Params *p, BITMAP *a, BITMAP *b, BITMAP *bmask=NULL, RegionMasks *region_masks=NULL, RegionMasks *amask=NULL, BITMAP **ann_sim=NULL);
BITMAP *vecwrap_init_dist(int vec_mode, Params *p, BITMAP *a, BITMAP *b, BITMAP *ann, BITMAP *bmask=NULL, RegionMasks *region_masks=NULL, RegionMasks *amask=NULL, BITMAP *ann_sim=NULL);
void vecwrap_nn(int vec_mode, Params *p, BITMAP *a, BITMAP *b,
        BITMAP *ann, BITMAP *annd,
        RegionMasks *amask=NULL, BITMAP *bmask=NULL,
        int level=0, int em_iter=0, RecomposeParams *rp=NULL, int offset_iter=0, int update_type=0, int cache_b=0,
        RegionMasks *region_masks=NULL, int tiles=-1, BITMAP *ann_sim=NULL);
BITMAP *vecwrap_vote(int vec_mode, Params *p, BITMAP *b,
             BITMAP *ann, BITMAP *ann_sim=NULL, BITMAP *bnn=NULL,
             BITMAP *bmask=NULL, BITMAP *bweight=NULL,
             double coherence_weight=COHERENCE_WEIGHT, double complete_weight=COMPLETE_WEIGHT,
             RegionMasks *amask=NULL, BITMAP *aweight=NULL, BITMAP *ainit=NULL, RegionMasks *region_masks=NULL, BITMAP *aconstraint=NULL, int mask_self_only=0);

#endif
